from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.pyplot as plt
import numpy as np
import typing
import xarray as xr
from collections.abc import MutableSet
import functools
import pathlib
import salvus.namespace as sn

from salvus.toolbox.helpers.wavefield_output import (
    WavefieldOutput,
    wavefield_output_to_xarray,
)


class OrderedSet(MutableSet):
    """
    An ordered set implementation using a dictionary for maintaining insertion
    order.

    Attributes:
        _items (dict): Internal dictionary to store set items in insertion
        order.
    """

    def __init__(self, iterable=None):
        """
        Initialize the OrderedSet.

        Args:
            iterable (iterable, optional): Optional initial values to populate the set.
        """
        self._items = dict()
        if iterable:
            self |= iterable

    def __len__(self):
        """Return the number of items in the set."""
        return len(self._items)

    def __iter__(self):
        """Return an iterator over the items in the set."""
        return iter(self._items)

    def __contains__(self, item):
        """Check if an item exists in the set."""
        return item in self._items

    def add(self, item):
        """Add an item to the set."""
        self._items[item] = None

    def discard(self, item):
        """Remove an item from the set if it exists."""
        self._items.pop(item, None)

    def __repr__(self):
        """Return a string representation of the set."""
        items = ", ".join(repr(item) for item in self._items)
        return f"{self.__class__.__name__}([{items}])"

    def __eq__(self, other):
        """Check equality with another OrderedSet."""
        if isinstance(other, OrderedSet):
            return list(self) == list(other)
        return False

    def __sub__(self, other):
        """Return the difference of two sets."""
        return OrderedSet(item for item in self if item not in other)

    def __and__(self, other):
        """Return the intersection of two sets."""
        return OrderedSet(item for item in self if item in other)

    def __or__(self, other):
        """Return the union of two sets."""
        return OrderedSet(
            list(self) + [item for item in other if item not in self]
        )

    def __xor__(self, other):
        """Return the symmetric difference of two sets."""
        return OrderedSet((self | other) - (self & other))


coordinates = OrderedSet(["x", "y", "z"])


# Helper functions
def get_wo(
    project,
    simulation_name: str,
    event: str,
    field: str,
    output_type: str = "volume",
):
    # Get wavefield output object
    output_dir = project.simulations.get_simulation_output_directory(
        simulation_configuration=simulation_name, event=event
    )

    return WavefieldOutput.from_file(
        output_dir / f"{output_type}_data_output.h5", field, output_type
    )


# Extract slices
def get_slice(
    project: sn.Project,
    simulation_name: str,
    event_name: str,
    output_field: str,
    coordinate: typing.Literal["x", "y", "z"],
    domain: sn.domain.Domain,
    resolution: int = 200,
):

    x0, x1 = domain.bounds.hc["x"]
    y0, y1 = domain.bounds.hc["y"]
    z0, z1 = domain.bounds.vc["z"]

    x_mid = (x0 + x1) / 2.0
    y_mid = (y0 + y1) / 2.0
    z_mid = (z0 + z1) / 2.0

    mids = [x_mid, y_mid, z_mid]
    selected_mid = mids[{"x": 0, "y": 1, "z": 2}.get(coordinate)]

    # Grid for the two others
    x_grid = np.linspace(x0, x1, resolution)
    y_grid = np.linspace(y0, y1, resolution)
    z_grid = np.linspace(z0, z1, resolution)

    # Full grid for wavefield_output_to_xarray, of which one will be set back
    # to a fixed point
    grid = [x_grid, y_grid, z_grid]
    grid[{"x": 0, "y": 1, "z": 2}.get(coordinate)] = np.array([selected_mid])

    return wavefield_output_to_xarray(
        get_wo(project, simulation_name, event_name, output_field), grid
    ).sel(**{coordinate: selected_mid})


def vector_norm(x, dim, ord=None):
    return xr.apply_ufunc(
        np.linalg.norm,
        x,
        input_core_dims=[[dim]],
        kwargs={"ord": ord, "axis": -1},
    )


def da_get_fixed_coordinates(da: xr.DataArray) -> set:
    """
    For DataArrays on which .sel or .isel has been applied,
    retrieve those coordinates that were selected.
    """
    return set(da.coords.keys()) - set(da.dims)


def plot_vector_magnitude(da, ax, label: str, unit: str):
    """
    Plot 3D vector data, stored along the `c` dimension in the DataArray, in
    the 2 present dimensions, stored in either x, y, or z.
    """

    # Ignore time (t), so we retrieve just the fixed slice coordinate.
    # Assumes no other coordinate has been selected.
    slice_coordinate = (da_get_fixed_coordinates(da) - {"t"}).pop()

    vector_magnitude = vector_norm(da.copy(), dim="c")

    extremum = np.max(vector_magnitude)

    img = vector_magnitude.T.plot(
        vmax=extremum,
        vmin=-extremum,
        cmap="seismic",
        ax=ax,
        # We'll do this manually.
        add_colorbar=False,
    )

    ax.set_aspect("equal")

    coord1, coord2 = coordinates - {slice_coordinate}

    title = (
        f"{coord1.upper()}-{coord2.upper()} plane, "
        f"{slice_coordinate} = "
        f"{da[slice_coordinate].values.item()}"
    )
    ax.set_title(title)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)

    cbar = plt.colorbar(img, cax=cax)

    cbar.set_label(f"{label} [{unit}]")


def plot_2d_vector_field(
    da: xr.DataArray,
    title: str | None = None,
    ax=None,
    global_max: float | None = None,
    max_resolution: int | None = None,
    scale: float = 1.0,
) -> None:
    """
    Plot the vector field for a ZY, YX, or ZX plane based on a sliced
    DataArray.

    """
    dataarray = da.copy()

    slice_coordinate = (da_get_fixed_coordinates(da) - {"t"}).pop()

    mapping = {"x": 0, "y": 1, "z": 2}

    varying_dims = set([dim for dim in dataarray.dims if dim != "c"])
    fixed_dim = set(mapping.keys()) - set(varying_dims)

    if len(varying_dims) != 2 or "c" not in dataarray.dims:
        raise ValueError(
            "DataArray must be sliced along one coordinate and have "
            "components `c=0, 1, 2`. Only one time slice can be passed."
        )

    coord1, coord2 = coordinates - fixed_dim

    ic1, ic2 = [mapping[vd] for vd in [coord1, coord2]]

    # Extract varying coordinates and displacement components
    coords1_values = dataarray[coord1].values
    coords2_values = dataarray[coord2].values
    component_1 = dataarray.sel(c=ic1).values
    component_2 = dataarray.sel(c=ic2).values

    # Subsample the data if max_resolution is provided
    if max_resolution is not None:
        stride1 = max(1, len(coords1_values) // max_resolution)
        stride2 = max(1, len(coords2_values) // max_resolution)
        coords1_values = coords1_values[::stride1]
        coords2_values = coords2_values[::stride2]
        component_1 = component_1[::stride1, ::stride2]
        component_2 = component_2[::stride1, ::stride2]

    # Create a meshgrid for quiver plot
    C1, C2 = np.meshgrid(coords1_values, coords2_values, indexing="ij")

    # Initialize Matplotlib OOP plot
    if ax is None:
        _, ax = plt.subplots(figsize=(10, 6))

    if global_max is None:
        global_max = np.max((component_1**2 + component_2**2) ** 0.5)

    # Create the quiver plot
    ax.quiver(
        C1,
        C2,
        component_1 / global_max,
        component_2 / global_max,
        angles="xy",
        scale_units="xy",
        scale=scale,
        color="blue",
    )

    # Generate a title if not provided
    if title is None:
        # title = f"Vector field ({coord1.upper()}-{coord2.upper()} Plane"
        title = f"{coord1.upper()}-{coord2.upper()} plane, {slice_coordinate} = {dataarray[slice_coordinate].values.item()}"

    # Add annotations
    ax.set_title(title, fontsize=14)
    ax.set_xlabel(f"{coord1.capitalize()} Coordinate", fontsize=12)
    ax.set_ylabel(f"{coord2.capitalize()} Coordinate", fontsize=12)
    # ax.grid(True)

    # Adjust axis limits for clarity
    ax.set_xlim(coords1_values.min(), coords1_values.max())
    ax.set_ylim(coords2_values.min(), coords2_values.max())

    # Set aspect ratio to square
    ax.set_aspect("equal", adjustable="box")

    return ax


def get_data(
    project: sn.Project,
    simulation: str,
    events: list[str],
    output_field: str,
    resolution: int = 200,
):
    return _cached_get_data(
        project._path, simulation, events, output_field, resolution
    )


# @functools.lru_cache(maxsize=32)
def _cached_get_data(
    project_path: pathlib.Path,
    simulation: str,
    events: list[str],
    output_field: str,
    resolution: int = 200,
):

    wo_iso_vpvs = {}

    project = sn.Project(project_path, verbosity=0)

    d = project.domain

    for event in events:
        wo_iso_vpvs[event] = {}
        for fixed_coordinate in coordinates:
            wo_iso_vpvs[event][fixed_coordinate] = get_slice(
                project,
                simulation,
                event_name=event,
                output_field=output_field,
                coordinate=fixed_coordinate,
                domain=d,
                resolution=resolution,
            )

    return wo_iso_vpvs


def visualize_magnitude(
    project: sn.Project,
    simulation: str,
    events: list[str],
    output_field: str,
    time: float,
    resolution: int = 200,
):

    wavefields = get_data(
        project,
        simulation,
        tuple(events),
        output_field,
        resolution=resolution,
    )

    for event in events:
        fig = plt.figure(figsize=(15, 5))
        for i_subplot, fixed_coordinate in enumerate(list(coordinates)[::-1]):

            ax = plt.subplot(1, 3, i_subplot + 1)

            wavefield_timeslice = wavefields[event][fixed_coordinate].sel(
                t=time, method="nearest"
            )

            plot_vector_magnitude(
                wavefield_timeslice, ax, label=output_field, unit="m"
            )

        fig.set_tight_layout(True)
        fig.suptitle(
            f"{output_field.capitalize()} magnitude for the "
            f"{event.capitalize()} event.\n"
            f"Time = {wavefield_timeslice.t.item():.2e} s"
        )


def visualize_vector(
    project: sn.Project,
    simulation: str,
    events: list[str],
    output_field: str,
    time: float,
    resolution: int = 200,
    max_vector_resolution: int = 50,
    scale: float = 1.0,
):

    wavefields = _cached_get_data(
        project._path,
        simulation,
        tuple(events),
        output_field,
        resolution=resolution,
    )

    for event in events:
        fig, axes = plt.subplots(1, 3, figsize=(13, 5))

        for ifc, fixed_coordinate in enumerate(list(coordinates)[::-1]):
            wavefield_timeslice = wavefields[event][fixed_coordinate].sel(
                t=time, method="nearest"
            )
            ax = axes[ifc]

            # Pass the manually set global_max
            plot_2d_vector_field(
                wavefield_timeslice,
                ax=ax,
                max_resolution=max_vector_resolution,
                scale=scale,
            )

        fig.set_tight_layout(True)
        fig.suptitle(
            f"{output_field.capitalize()} vector for the "
            f"{event.capitalize()} event.\n"
            f"Time = {wavefield_timeslice.t.item():.2e} s"
        )
